/*****************************************************************************
*
*	Collection of headers
*
*	Osamu Gotoh, ph.D.	(-2001)
*	Saitama Cancer Center Research Institute
*	818 Komuro, Ina-machi, Saitama 362-0806, Japan
*
*	Osamu Gotoh, Ph.D.	(2001-)
*	National Institute of Advanced Industrial Science and Technology
*	Computational Biology Research Center (CBRC)
*	2-41-6 Aomi, Koutou-ku, Tokyo 135-0064, Japan
*
*	Osamu Gotoh, Ph.D.      (2003-)
*	Department of Intelligence Science and Technology
*	Graduate School of Informatics, Kyoto University
*	Yoshida Honmachi, Sakyo-ku, Kyoto 606-8501, Japan
*
*	Copyright(c) Osamu Gotoh <<o.gotoh@aist.go.jp>>
*****************************************************************************/

#ifndef _STDTYPE_
#define _STDTYPE_

#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>
#include <time.h>
#include <float.h>
/*#include <err.h>*/
/*#include <values.h>*/
#include "adddef.h"
#include "clib.h"
#include "iolib.h"
#include "mfile.h"
#include "sets.h"

#endif
